from sqlalchemy.exc import InterfaceError
import threading


_local = threading.local()

key = 'SM'
skey = 'SM.sessions'

_registered_sessions = {}


def _getRegisteredSessions():
    return _registered_sessions
    #if not hasattr(_local, 'data'):
    #    _local.data = {}
    #return _local.data


def registerSession(name, Session):
    _getRegisteredSessions()[name] = Session


def getSessionManager(environ):
    return environ[key]


class SM(object):
    """
    yes, I know this is sort of wonky...

    Sooo.. what's the point?

    Well, I have apps that I write which have multiple factored instances
    and I need a way to make them all our together okay.

    So this allows me to jungle the different app sessions.
    """

    def __init__(self, environ):
        self.environ = environ
        if skey not in environ:
            environ[skey] = {}
        self.sessions = environ[skey]

    def __getattr__(self, name):
        if name not in _getRegisteredSessions():
            raise AttributeError
        if name not in self.sessions:
            session = _getRegisteredSessions()[name]
            if callable(session):
                session = session()
            self.sessions[name] = session
        return self.sessions[name]

    __getitem__ = __getattr__

    def commit(self):
        for session in self.sessions.values():
            if hasattr(session, 'commit'):
                session.commit()

    def rollback(self):
        for session in self.sessions.values():
            try:
                if hasattr(session, 'rollback'):
                    session.rollback()
            except InterfaceError:
                pass

    def close(self):
        for session in self.sessions.values():
            if hasattr('session', 'close'):
                session.close()


class SMFilter(object):
    def __init__(self, application):
        self.application = application

    def __call__(self, environ, start_response):
        if key not in environ:
            environ[key] = SM(environ)
        sm = environ[key]

        try:
            result = self.application(environ, start_response)
            sm.commit()
        except:
            sm.rollback()
            raise

        sm.close()
        if key in environ:
            del environ[key]
        if skey in environ:
            del environ[skey]
        return result


def make_sm(app, global_conf):
    return SMFilter(app)



from sqlalchemy import Column, Integer, Text, DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from factored.sm import registerSession
from datetime import datetime
from factored.utils import make_random_code


DBSession = sessionmaker()
registerSession('f', DBSession)
Base = declarative_base()


def generate_user_code():
    return make_random_code(12)


class User(Base):
    __tablename__ = 'users'
    id = Column(Integer, primary_key=True)
    username = Column(Text, unique=True)
    secret = Column(Text, unique=True)
    generated_code = Column(Text, default=generate_user_code)
    generated_code_time_stamp = Column(DateTime, default=datetime.utcnow)

    def __init__(self, username, secret):
        self.username = username
        self.secret = secret



import os
import sys

from sqlalchemy import engine_from_config
from pyramid.paster import get_appsettings, setup_logging

from factored.models import DBSession, Base


def usage(argv):
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri>\n'
          '(example: "%s development.ini")' % (cmd, cmd))
    sys.exit(1)


def main(argv=sys.argv):
    if len(argv) != 2:
        usage(argv)
    config_uri = argv[1]
    setup_logging(config_uri)
    try:
        settings = get_appsettings(config_uri, 'factored')
    except LookupError:
        settings = get_appsettings(config_uri, 'main')
    engine = engine_from_config(settings, 'sqlalchemy.')
    DBSession.configure(bind=engine)
    Base.metadata.create_all(engine)



import os
from pyramid.httpexceptions import HTTPFound
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode
from factored.plugins import getFactoredPlugin, getFactoredPlugins
from factored.plugins import getPluginForRequest
from factored.utils import generate_url


class AuthView(object):

    def __init__(self, req):
        self.req = req
        self.plugin = getPluginForRequest(req)

    def __call__(self):
        return self.plugin.render()


def notfound(req):
    base_path = req.registry['settings']['base_auth_url']
    orig_url = generate_url(req, req.path)
    redirect_url = generate_url(
        req,
        base_path,
        params=dict(referrer=orig_url))
    return HTTPFound(location=redirect_url)


def auth_chooser(req):
    auth_types = []
    settings = req.registry['settings']
    base_path = settings['base_auth_url']
    supported_types = settings['supported_auth_schemes']
    if len(supported_types) == 1:
        plugin = getFactoredPlugin(supported_types[0])
        url = generate_url(req, base_path + '/' + plugin.path,
                           {'referrer': req.params.get('referrer', '')})
        raise HTTPFound(location=url)

    for plugin in getFactoredPlugins():
        if plugin.name in supported_types:
            auth_types.append({
                'name': plugin.name,
                'url': os.path.join(base_path, plugin.path)
                })
    return dict(auth_types=auth_types,
                referrer=req.params.get('referrer', ''))



from pyramid.config import Configurator
from factored.auth_tkt import AuthTktAuthenticator, AuthenticationPolicy
from factored.finders import getUserFinderPlugin
from factored import TEMPLATE_CUSTOMIZATIONS
import os
from pyramid_mailer.mailer import Mailer
from factored import subscribers
subscribers  # pyflakes
from factored.request import Request
from factored.sm import SMFilter

from pkg_resources import iter_entry_points


def _tolist(val):
    if type(val) == list:
        return val
    lines = val.splitlines()
    return [l.strip() for l in lines if l.strip()]


def get_settings(config, prefix):
    settings = {}
    for key, val in config.items():
        if key.startswith(prefix):
            settings[key[len(prefix):]] = val
    return settings


def nested_settings(settings):
    newsettings = {}
    for key, val in settings.items():
        parts = key.split('.')
        loc = newsettings
        for part in parts[:-1]:
            if part not in loc:
                loc[part] = {}
            loc = loc[part]
        loc[parts[-1]] = val
    return newsettings


def normalize_settings(settings):
    for key, val in settings.items():
        if val in ('true', 'True', 't'):
            settings[key] = True
        elif val in ('false', 'False', 'f'):
            settings[key] = False
    return settings


db_factories = dict([(i.name, i.load()) for i in iter_entry_points(
                     group='factored.db_factory', name=None)])


class Authenticator(object):

    def __init__(self, *args, **settings):
        if len(args) == 1:
            # regular app
            app = None
        else:
            app = args[0]
        self.initialize_settings(app, settings)

        # db configuration
        db_factory_name = settings.get('db_factory_name', 'sql')
        if db_factory_name not in db_factories:
            raise Exception("Invalid db_factory_name: %s" % db_factory_name)
        settings['db_session_id'], self.db = db_factories[db_factory_name](settings, self)  # noqa

        self.setup_autouserfinder(settings)

        # start pyramid application configuration
        config = Configurator(settings=settings, request_factory=Request)
        try:
            import pyramid_chameleon  # noqa
            config.include('pyramid_chameleon')
        except ImportError:
            pass

        self.setup_plugins(config, settings)

        from factored.views import auth_chooser, notfound
        config.add_route('auth', self.base_auth_url)
        config.add_view(auth_chooser, route_name='auth',
                        renderer='templates/layout.pt')

        # setup template customization registration
        if TEMPLATE_CUSTOMIZATIONS not in config.registry:
            config.registry[TEMPLATE_CUSTOMIZATIONS] = {}

        # static paths for resources
        self.static_path = os.path.join(self.base_auth_url, 'authstatic')
        config.add_static_view(name=self.static_path,
                               path='factored:static')
        config.add_notfound_view(notfound, append_slash=True)

        # add some things to registry
        config.registry['mailer'] = Mailer.from_settings(settings)
        config.registry['settings'] = self.__dict__
        config.registry['formtext'] = nested_settings(
            get_settings(settings, 'formtext.'))
        config.registry['app'] = self

        config.scan()
        self.config = config
        self.registry = self.config.registry
        try:
            self.app.config.registry['factored'] = self
        except:
            pass
        self.pyramid = config.make_wsgi_app()

    def setup_plugins(self, config, settings):
        from factored.plugins import getFactoredPlugins
        from factored.views import AuthView
        for plugin in getFactoredPlugins():
            setattr(
                self, '%s_settings' % plugin.path,
                nested_settings(get_settings(settings, '%s.' % plugin.path)))
            config.add_route(
                plugin.name,
                os.path.join(self.base_auth_url, plugin.path))
            config.add_view(AuthView, route_name=plugin.name,
                            renderer='templates/layout.pt')

    def setup_autouserfinder(self, settings):
        finder_name = settings.get('autouserfinder', None)
        if finder_name:
            plugin = getUserFinderPlugin(finder_name)
            if plugin:
                self.userfinder = plugin(
                    **get_settings(settings, 'autouserfinder.'))
            else:
                raise Exception('User finder not found: %s', finder_name)

    def initialize_settings(self, app, settings):
        self.app = app
        self.appname = settings.pop('appname', 'REPLACEME')
        self.supported_auth_schemes = _tolist(
            settings.pop('supported_auth_schemes', "Google Auth"))
        self.base_auth_url = settings.pop('base_auth_url', '/auth')
        self.email_auth_window = int(settings.pop('email_auth_window', '120'))
        self.auth_timeout = int(settings.pop('auth_timeout', '7200'))
        self.auth_remember_timeout = int(settings.pop(
            'auth_remember_timeout', '86400'))

        auth_settings = normalize_settings(get_settings(settings, 'auth_tkt.'))
        self.auth_tkt_policy = AuthenticationPolicy(**auth_settings)
        self.allowcodereminder = settings.pop(
            'allowcodereminder', 'false').lower() == 'true' or False
        self.allowcodereminder_settings = get_settings(
            settings, 'allowcodereminder.')

        settings['hide_banner'] = settings.get(
            'hide_banner', 'false').strip().lower() == 'true'
        self.hide_banner = settings['hide_banner']

        self.excepted_paths = [a.strip() for a in
            settings.get("excepted_paths", "").splitlines()]

    def __call__(self, environ, start_response):
        def wrapped_app(environ2, start_response2):
            auth = AuthTktAuthenticator(self.auth_tkt_policy, environ2)
            environ2['auth'] = auth
            path = environ2['PATH_INFO']
            excepted_paths = self.excepted_paths
            auth_result = auth.authenticate()
            if path in excepted_paths or auth_result:
                if self.app is not None:
                    # if this is a filter, we can pass on to the actual app
                    return self.app(environ2, start_response2)
            return self.pyramid(environ2, start_response2)
        return SMFilter(wrapped_app)(environ, start_response)


try:
    from wsgiproxy.exactproxy import proxy_exact_request

    class SimpleProxy(object):

        def __init__(self, global_config, server, port, urlscheme=None):
            self.server = server
            self.port = port
            self.scheme = urlscheme

        def __call__(self, environ, start_response):
            environ['SERVER_NAME'] = self.server
            environ['SERVER_PORT'] = self.port
            if self.scheme:
                environ['wsgi.url_scheme'] = self.scheme
            return proxy_exact_request(environ, start_response)
except ImportError:
    class SimpleProxy(object):
        def __init__(self, *args, **kwargs):
            raise Exception("Must install factored with [proxy] "
                            "requirement to use this.")



from pyramid.renderers import render
from factored import TEMPLATE_CUSTOMIZATIONS
import os
from pyramid.path import package_path


def registerTemplateCustomizations(config, _dir, package):
    registry = config.registry
    if TEMPLATE_CUSTOMIZATIONS not in registry:
        registry[TEMPLATE_CUSTOMIZATIONS] = {}
    path = os.path.join(package_path(package), _dir)
    for fi in os.listdir(path):
        registry[TEMPLATE_CUSTOMIZATIONS][fi] = (
            package, os.path.join(_dir, fi))


class TemplateRendererFactory(object):

    def __init__(self, req, context):
        self.context = context
        self.req = req
        self.customizations = req.registry[TEMPLATE_CUSTOMIZATIONS]
        if not self.customizations and 'app' in req.registry:
            try:
                self.customizations = self.req.registry['app'].app.config.registry[TEMPLATE_CUSTOMIZATIONS]  # noqa
            except:
                pass

    def render(self, tmpl, package=None):
        try:
            tmpl_name = os.path.basename(tmpl)
            if tmpl_name in self.customizations:
                package, tmpl = self.customizations[tmpl_name]
            if package is None and ':' in tmpl:
                pkg, tmpl = tmpl.split(':', 1)
                try:
                    package = __import__(pkg)
                except ImportError:
                    pass
            return render(tmpl, self.context, request=self.req,
                          package=package)
        except ValueError:
            return ''



import base64
from datetime import datetime
from hashlib import sha256 as sha
import hashlib
import hmac
import struct
import time

from pyramid.threadlocal import get_current_registry

try:
    from urllib import urlencode
except:
    from urllib.parse import urlencode

import random
try:
    random = random.SystemRandom()
    using_sysrandom = True
except NotImplementedError:
    using_sysrandom = False


# generated when process started, hard to guess
SECRET = random.randint(0, 1000000)


def get_random_string(length=12,
                      allowed_chars='abcdefghijklmnopqrstuvwxyz'
                                    'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'):
    """
    Returns a securely generated random string.

    The default length of 12 with the a-z, A-Z, 0-9 character set returns
    a 71-bit value. log_2((26+26+10)^12) =~ 71 bits
    """
    if not using_sysrandom:
        # This is ugly, and a hack, but it makes things better than
        # the alternative of predictability. This re-seeds the PRNG
        # using a value that is hard for an attacker to predict, every
        # time a random string is required. This may change the
        # properties of the chosen random sequence slightly, but this
        # is better than absolute predictability.
        random.seed(
            sha(
                "%s%s%s" % (
                    random.getstate(),
                    time.time(),
                    SECRET)
                ).digest())
    return ''.join([random.choice(allowed_chars) for i in range(length)])


def generate_random_google_code(length=10):
    return base64.b32encode(get_random_string(length))


def make_random_code(length=255):
    prehash = hashlib.sha1(str(get_random_string(length)).encode('utf-8')).hexdigest()[:5]
    return hashlib.sha1(
        (prehash + str(datetime.now().microsecond)).encode('utf-8')).hexdigest()[:length]


def get_barcode_image(username, secretkey, appname):
    url = "https://www.google.com/chart"
    url += "?chs=200x200&chld=M|0&cht=qr&chl=otpauth://totp/"
    username = username + '--' + appname
    url += username + "%3Fsecret%3D" + secretkey
    return url


def get_google_auth_code(secretkey, tm=None):
    if tm is None:
        tm = int(time.time() / 30)
    secretkey = base64.b32decode(secretkey)
    # convert timestamp to raw bytes
    b = struct.pack(">q", tm)

    # generate HMAC-SHA1 from timestamp based on secret key
    hm = hmac.HMAC(secretkey, b, hashlib.sha1).digest()

    # extract 4 bytes from digest based on LSB
    offset = ord(hm[-1]) & 0x0F
    truncatedHash = hm[offset:offset + 4]

    # get the code from it
    code = struct.unpack(">L", truncatedHash)[0]
    code &= 0x7FFFFFFF
    code %= 1000000
    return "%06d" % code


def create_user(username, session=None):
    from factored.models import DBSession, User
    if session is None:
        session = DBSession()
    secret = generate_random_google_code()
    user = User(username=username, secret=secret)
    session.add(user)
    return user


NO_VALUE = object()


class CombinedDict(object):
    def __init__(self, *args):
        self.dicts = args

    def __getattr__(self, name, default=NO_VALUE):
        try:
            return self[name]
        except KeyError:
            if NO_VALUE == default:
                raise AttributeError(name)
            return default

    def __getitem__(self, name):
        """
        need to handle nested dictionaries also
        """
        founddicts = []
        for dic in self.dicts:
            if name in dic:
                val = dic[name]
                if type(val) == dict:
                    founddicts.append(val)
                else:
                    return val
        if founddicts:
            return CombinedDict(*founddicts)
        raise KeyError(name)


class FakeMailer(object):

    def __init__(self, req):
        self.req = req
        self.messages = []

    def send_immediately(self, message):
        self.messages.append(message)
        print("""To: %s
From: %s
Subject: %s
Body: %s""" % (
            ','.join(message.recipients),
            message.sender,
            message.subject,
            message.body
        ))

    send = send_immediately


FAKE_MAILER_KEY = 'fakemailer'


def get_mailer(req):
    settings = req.registry.settings
    if settings.get('mail.host', '') == 'debug':
        if FAKE_MAILER_KEY not in req.environ:
            req.environ[FAKE_MAILER_KEY] = FakeMailer(req)
        return req.environ[FAKE_MAILER_KEY]
    return req.registry['mailer']


def generate_url(req, path, params={}):
    scheme = None
    if 'X-Forwarded-Protocol' in req.headers:
        scheme = req.headers['X-Forwarded-Protocol']
    elif 'wsgi.url_scheme' in req.environ:
        scheme = req.environ['wsgi.url_scheme']
    elif req.url.startswith('https://'):
        scheme = 'https'
    if scheme is None:
        # look for HTTP_ORIGIN
        if 'HTTP_ORIGIN' in req.environ:
            host = req.environ['HTTP_ORIGIN']
        else:
            host = ''
    else:
        host = '%s://%s' % (scheme, req.environ['HTTP_HOST'])
    url = '%s%s' % (host, path)
    if params:
        url += '?' + urlencode(params)
    return url


def create_message_id(_id=''):
    registry = get_current_registry()
    domain = 'localhost'
    if registry:
        domain = registry.settings.get('mail.domain', domain)
    if not _id:
        _id = '%s-%s' % (
            str(time.time()),
            get_random_string(20))
    return '<%s@%s>' % (_id, domain)


from webob.request import BaseRequest as Request
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authentication import AuthTktCookieHelper
try:
    from pyramid.authentication import EXPIRE
except ImportError:
    EXPIRE = object()
import datetime

_marker = object()


class CookieHelper(AuthTktCookieHelper):
    def __init__(self, *args, **kwargs):
        self.cookie_domain = kwargs.pop('cookie_domain', False)
        super(CookieHelper, self).__init__(*args, **kwargs)

    def _get_cookies(self, environ, value, max_age=None):
        cookies = super(CookieHelper, self)._get_cookies(environ, value,
                                                         max_age)
        if self.cookie_domain:
            if max_age is EXPIRE:
                max_age = "; Max-Age=0; Expires=Wed, 31-Dec-97 23:59:59 GMT"
            elif max_age is not None:
                later = datetime.datetime.utcnow() + datetime.timedelta(
                    seconds=int(max_age))
                # Wdy, DD-Mon-YY HH:MM:SS GMT
                expires = later.strftime('%a, %d %b %Y %H:%M:%S GMT')
                # the Expires header is *required* at least for IE7 (IE7 does
                # not respect Max-Age)
                max_age = "; Max-Age=%s; Expires=%s" % (max_age, expires)
            else:
                max_age = ''

            cookies.append(('Set-Cookie', '%s="%s"; Path=%s; Domain=%s%s%s' % (
                self.cookie_name, value, self.path,
                self.cookie_domain, max_age, self.static_flags)))
        return cookies


class AuthenticationPolicy(AuthTktAuthenticationPolicy):
    def __init__(self, secret, callback=None, cookie_name='auth_tkt',
                 secure=False, include_ip=False, timeout=None,
                 reissue_time=None, max_age=None, path="/", http_only=False,
                 wild_domain=True, cookie_domain=False, debug=False,
                 hashalg='sha512'):
        self.cookie = CookieHelper(
            secret,
            cookie_name=cookie_name,
            secure=secure,
            include_ip=include_ip,
            timeout=timeout,
            reissue_time=reissue_time,
            max_age=max_age,
            http_only=http_only,
            path=path,
            wild_domain=wild_domain,
            cookie_domain=cookie_domain,
            hashalg=hashalg)
        self.callback = callback
        self.debug = debug


class AuthTktAuthenticator(object):

    def __init__(self, policy, environ):
        self.policy = policy
        self.environ = environ
        self.request = Request(environ)

    def remember(self, principal, **kw):
        return self.policy.remember(self.request, principal, **kw)

    def authenticate(self):
        identity = self.policy.unauthenticated_userid(self.request)
        if identity:
            self.environ['factored.identity'] = identity
            # set the REMOTE_USER
            self.environ['REMOTE_USER'] = identity
            # also set a header for username
            self.environ['HTTP_FACTORED_USER'] = identity
            return identity



from copy import deepcopy as copy
from datetime import datetime
from datetime import timedelta
from formencode import Schema, validators
import hashlib
import os
from pyramid.httpexceptions import HTTPFound, HTTPInternalServerError
from pyramid.util import strings_differ
from pyramid_mailer.message import Message
from pyramid_simpleform import Form
from pyramid_simpleform.renderers import FormRenderer
import requests
import string
import time
try:
    from urlparse import urlparse, parse_qsl
except ImportError:
    from urllib.parse import urlparse, parse_qsl

from factored.utils import CombinedDict
from factored.utils import get_barcode_image
from factored.utils import get_google_auth_code
from factored.utils import get_mailer
from factored.utils import make_random_code
from factored.utils import generate_url
from factored.utils import create_message_id

import logging
logger = logging.getLogger(__name__)


_auth_plugins = []


def safe_headers(values):
    headers = []
    for name, value in values:
        if isinstance(value, unicode):
            value = value.encode('utf8')
        headers.append((name, value))
    return headers


def nested_update(base, override):
    for key, val in override.items():
        baseval = base.get(key)
        if type(baseval) == dict:
            nested_update(baseval, val)
        else:
            base[key] = val
    return base


class BaseSchema(Schema):
    filter_extra_fields = True
    allow_extra_fields = True


def addFactoredPlugin(plugin):
    _auth_plugins.append(plugin)


def getFactoredPlugins():
    return _auth_plugins


def getFactoredPlugin(name):
    for plugin in _auth_plugins:
        if plugin.name == name:
            return plugin


def getPluginForRequest(req):
    app = req.registry['app']
    for plugin in getFactoredPlugins():
        pluginpath = os.path.join(app.base_auth_url, plugin.path)
        if pluginpath == req.path:
            return plugin(req)


class UsernameSchema(BaseSchema):
    username = validators.MinLength(3, not_empty=True)
    referrer = validators.UnicodeString(if_missing=u'')


class CodeSchema(BaseSchema):
    username = validators.MinLength(3, not_empty=True)
    code = validators.MinLength(4, not_empty=True)
    remember = validators.Bool()
    referrer = validators.UnicodeString(if_missing=u'')


class BasePlugin(object):
    name = None
    path = None
    username_schema = UsernameSchema
    code_schema = CodeSchema
    content_renderer = "templates/auth.pt"

    _formtext = {
        'title': None,
        'legend': None,
        'username': {
            'label': None,
            'desc': None
        },
        'code': {
            'label': u'Access Code',
            'desc': None
        },
        'button': {
            'username': u'Next',
            'authenticate': u'Authenticate',
            'codereminder': u'Send Code Reminder'
        },
        'error': {
            'invalid_username_code': u'Invalid username for this code',
            'invalid_code': u'Code did not validate',
            'invalid_username': u'Invalid username',
            'code_reminder': u'A code reminder email has been sent.'
        }
    }
    _formtext_overrides = {}

    def __init__(self, req):
        self.req = req
        self.uform = Form(req, schema=self.username_schema)
        self.cform = Form(req, schema=self.code_schema)
        self.formtext = nested_update(copy(self._formtext),
                                      self._formtext_overrides)
        self.send_submitted = self.validate_submitted = False
        self.combined_formtext = CombinedDict(
            req.registry['formtext'], self.formtext)

        self.auth_timeout = req.registry['settings']['auth_timeout']
        self.auth_remember_timeout = \
            req.registry['settings']['auth_remember_timeout']
        rto = self.auth_remember_timeout / 60
        if rto > 60:
            rto = rto / 60
            if rto > 24:
                rto = '%i days' % (rto / 24)
            else:
                rto = '%i hours' % rto
        else:
            rto = '%i minutes' % rto
        self.remember_duration = rto

    @property
    def app(self):
        return self.req.registry['app']

    @property
    def db(self):
        return self.app.db

    def get_user(self, username):
        user = self.db.get_user(self.req, username.lower())
        if user is None:
            if 'userfinder' in self.req.registry['settings']:
                finder = self.req.registry['settings']['userfinder']
                if finder(username):
                    return self.db.create_user(self.req, username)
        return user

    def check_code(self, user):
        return False

    def user_form_submitted_successfully(self, user):
        pass

    @property
    def allow_code_reminder(self):
        return False

    def send_code_reminder(self, user):
        pass

    @property
    def settings(self):
        name = '%s_settings' % self.path
        if name in self.req.registry['settings']:
            return self.req.registry['settings'][name]
        else:
            return {}

    def update_referrer(self):
        referrer = self.uform.data.get('referrer',
                                       self.req.params.get('referrer', ''))
        referrer = self.cform.data.get('referrer', referrer)
        if not referrer:
            referrer = self.req.path_url
        # strip off auth url
        parts = self.req.environ['PATH_INFO'].rsplit('/')
        referrer = referrer.rstrip('/')
        if parts[-1] == self.path and \
                referrer.endswith(self.path):
            referrer = referrer[:-len(self.path)]
            parts.remove(self.path)
        referrer = referrer.rstrip('/')
        if referrer.endswith(parts[-1]):
            referrer = referrer[:-len(parts[-1])]
        self.uform.data.update({'referrer': referrer})
        self.cform.data.update({'referrer': referrer})

    @property
    def requesting_code_reminder(self):
        cm = self.combined_formtext['button']['codereminder']
        acm = self.allow_code_reminder
        return self.req.POST.get('submit', '') == cm and acm

    def send_out_code_reminder(self):
        self.validate_submitted = True
        self.cform.validate()
        self.cform.errors['code'] = self.error_code_reminder
        username = self.cform.data['username'].lower()
        user = self.get_user(username)
        if 'username' not in self.cform.errors and user is not None:
            self.send_code_reminder(user)

    @property
    def requesting_user_form(self):
        return self.req.POST.get('submit', '') == \
            self.combined_formtext['button']['username']

    def submit_user_form(self):
        if self.uform.validate():
            username = self.uform.data['username'].lower()
            self.cform.data['username'] = username
            user = self.get_user(username)
            if user is None:
                self.uform.errors['username'] = \
                    self.combined_formtext['error']['invalid_username']
            else:
                self.send_submitted = True
                self.user_form_submitted_successfully(user)

    @property
    def requesting_authentication(self):
        return self.req.POST.get('submit', '') == \
            self.combined_formtext['button']['authenticate']

    def submit_authentication(self):
        self.validate_submitted = True
        if self.cform.validate():
            user = self.get_user(self.cform.data['username'].lower())
            if user is None:
                self.cform.errors['code'] = \
                    self.combined_formtext['error']['invalid_username_code']
            else:
                if self.check_code(user):
                    userid = self.cform.data['username'].lower()
                    if self.cform.data['remember']:
                        max_age = self.auth_remember_timeout
                    else:
                        max_age = self.auth_timeout
                    auth = self.req.environ['auth']
                    headers = safe_headers(auth.remember(userid, max_age=max_age))  # noqa
                    referrer = self.cform.data.get('referrer')
                    if not referrer:
                        referrer = '/'
                    raise HTTPFound(location=referrer, headers=headers)
                else:
                    self.cform.errors['code'] = \
                        self.combined_formtext['error']['invalid_code']
                    self.cform.data['code'] = u''

    def render(self):
        req = self.req
        if req.method == "POST":
            if self.requesting_code_reminder:
                self.send_out_code_reminder()
            elif self.requesting_user_form:
                self.submit_user_form()
            elif self.requesting_authentication:
                self.submit_authentication()
        self.update_referrer()
        return dict(
            auth_plugin=self,
            uform=FormRenderer(self.uform),
            cform=FormRenderer(self.cform),
            send_submitted=self.send_submitted,
            validate_submitted=self.validate_submitted,
            content_renderer=self.content_renderer,
            formtext=self.combined_formtext,
            remember_duration=self.remember_duration)


class GoogleAuthPlugin(BasePlugin):
    name = 'Google Auth'
    path = 'ga'

    _formtext_overrides = {
        'title': u'Google Authenticator',
        'legend': u'Use android app to authenticate...',
        'username': {
            'label': u'Username',
            'desc': u"Username you've signed up with."
        },
        'code': {
            'desc': u'As generated with google authenticator.'
        }
    }

    @property
    def allow_code_reminder(self):
        return self.req.registry['settings']['allowcodereminder']

    def send_code_reminder(self, user):
        mailer = get_mailer(self.req)
        code_reminder_settings = \
            self.req.registry['settings']['allowcodereminder_settings'].copy()
        username = user.username
        message = code_reminder_settings.copy()
        message['recipients'] = [username]
        appname = self.req.registry['settings']['appname']
        message['body'] = message['body'].replace(
            '{code}', get_barcode_image(username, user.secret, appname))
        message['extra_headers'] = {
            'Message-ID': create_message_id()
        }
        mailer.send_immediately(Message(**message))

    def check_code(self, user):
        tm = int(time.time() / 30)
        code_attempt = self.cform.data['code']
        # try 30 seconds behind and ahead as well
        for ix in [-1, 0, 1]:
            code = get_google_auth_code(user.secret, tm + ix)

            if not strings_differ(code, str(code_attempt)):
                return True
        return False


addFactoredPlugin(GoogleAuthPlugin)


class EmailAuthSchema(UsernameSchema):
    username = validators.Email(not_empty=True)


class EmailAuthCodeSchema(CodeSchema):
    username = validators.Email(not_empty=True)
    code = validators.MinLength(8, not_empty=True)


class EmailAuthPlugin(BasePlugin):
    name = 'Email'
    path = 'em'
    username_schema = EmailAuthSchema
    code_schema = EmailAuthCodeSchema

    __formtext_overrides = {
        'title': u'Email Authenticator',
        'legend': u'Authenticate through your email...',
        'username': {
            'label': u'Email',
            'desc': u'Email you signed up with. Should be '
                    u'the same as the username.'
        },
        'code': {
            'desc': u'Provided in the email sent to you.'
        },
        'button': {
            'username': u'Send mail'
        }
    }

    @property
    def _formtext_overrides(self):
        overrides = self.__formtext_overrides.copy()
        if "{url}" in self.settings['body']:
            overrides['code']['desc'] = ('Provided in email sent to you in '
                                         'addition to an email link to login '
                                         'with automatically.')
        else:
            overrides['code']['desc'] = 'Provided in email sent to you.'
        return overrides

    def __init__(self, req):
        super(EmailAuthPlugin, self).__init__(req)

    def submit_authentication(self):
        urlcode = self.req.GET.get('code', None)
        if urlcode:
            referrer = self.req.params.get('referrer', '/')
            remember = self.req.params.get('rem', '0')
            remember = remember != '0'
            urlname = self.req.params.get('u', None)
            user = self.get_user(urlname.lower())
            # make sure the passed user is valid
            if user is None:
                self.cform.errors['code'] = \
                    self.combined_formtext['error']['invalid_username_code']
                return
            # make sure the urlcode hash matches with the user id of the user
            #   identified from the url, generated code, and salt
            code = user.generated_code
            salt = self.settings.get('salt', 's4lt!necracker')
            newhash = hashlib.sha256(
                "%s%s%s" % (user.id, code, salt)).hexdigest()
            if newhash != urlcode:
                self.cform.errors['code'] = \
                    self.combined_formtext['error']['invalid_code']
                return
            # check the code (make sure it's still valid and all that)
            #   note: even though this is comparing the same actual code value
            #   to itself, it's still checking to see if the code timed out,
            #   etc
            if self.check_code(user, user.generated_code):
                userid = user.username
                if remember:
                    max_age = self.auth_remember_timeout
                else:
                    max_age = self.auth_timeout
                auth = self.req.environ['auth']
                headers = safe_headers(auth.remember(userid, max_age=max_age))
                raise HTTPFound(location=referrer, headers=headers)
            else:
                self.cform.errors['code'] = \
                    self.combined_formtext['error']['invalid_code']
                self.cform.data['code'] = u''
                return
        else:
            super(EmailAuthPlugin, self).submit_authentication()

    def user_form_submitted_successfully(self, user):
        username = self.uform.data['username']
        mailer = get_mailer(self.req)
        user.generated_code = make_random_code(12)
        user.generated_code_time_stamp = datetime.utcnow()
        # save it
        self.db.save(self.req, user)
        settings = self.settings

        # only generate the url if it's being used
        url = ''
        if "{url}" in settings['body']:
            salt = settings.get('salt', 's4lt!necracker')
            urlparts = list(urlparse(self.req.url))
            querystr = dict(parse_qsl(urlparts[4]))
            querystr['code'] = hashlib.sha256(
                "%s%s%s" % (user.id, user.generated_code, salt)) \
                .hexdigest()
            querystr['u'] = username
            querystr['rem'] = settings.get('url_remember', '0')
            querystr['rem'] = '1' if querystr['rem'] == 'True' else '0'
            url = generate_url(self.req, self.req.path, querystr)

        message = {
            'recipients': [username],
            'body': settings['body'].replace('{code}', user.generated_code)
                                    .replace('{url}', url),
            'subject': settings['subject'],
            'sender': settings['sender'],
            'extra_headers': {
                'Message-ID': create_message_id()
            }
        }
        mailer.send_immediately(Message(**message))

    def check_code(self, user, code=None):
        window = self.req.registry['settings']['email_auth_window']
        now = datetime.utcnow()
        codetocheck = code if code else self.cform.data['code']
        return ((not strings_differ(codetocheck, user.generated_code))
                and (now < (user.generated_code_time_stamp +
                            timedelta(seconds=window))))

    def render(self):
        if self.req.GET.get('code', None):
            self.validate_submitted = True
            self.submit_authentication()

        return super(EmailAuthPlugin, self).render()


addFactoredPlugin(EmailAuthPlugin)


class SMSAuthPlugin(BasePlugin):
    name = "SMS"
    path = "sms"

    _formtext_overrides = {
        'title': u'SMS Authenticator',
        'legend': u'Authenticate through an sms message...',
        'username': {
            'label': u'Username/EMail',
            'desc': u'Your username (often an EMail address)'
        },
        'code': {
            'desc': u'Provided in the text message sent to your phone.'
        },
        'button': {
            'username': u'Send SMS'
        }
    }

    def get_user(self, username):
        user = super(SMSAuthPlugin, self).get_user(username)
        
        # SMS requires a phone number, but autouserfinders have no (current)
        # way of getting more information about a user than the username.
        # The best recourse, without bumpining into a major version update,
        # for now, is to reach out to an api that will give us a phone number
        # for a given username. The auth remains decoupled from the user finder
        # for now. Ideally the user finders would be able to return a configurable
        # amount of data, some of which could be phone numbers -- but that's
        # for a future update ;)

        endpoint = self.req.registry.settings.get("sms.userlist", None)
        if not endpoint:
            logger.error("sms.userlist endpoint not configured")
            raise HTTPInternalServerError()
        if "{username}" not in endpoint:
            logger.error("sms.userlist endpoint not configured with {username} variable")
            raise HTTPInternalServerError()

        endpoint = endpoint.format(username=username)
        try:
            response = requests.get(endpoint)
            response.raise_for_status()
            user.phone = response.json()["phone"]
        except requests.exceptions.HTTPError as ex:
            logger.error("error getting user info from sms.userlist endpoint ({} -- {}):".format(
                ex.response.status_code,
                ex.response.text))
            logger.error(ex)
            raise HTTPInternalServerError()
        except ValueError as ex:
            logger.error(ex)
            raise HTTPInternalServerError()

        return user

    def user_form_submitted_successfully(self, user):
        # grab configuration
        sms_msg_tmpl = self.req.registry.settings.get(
            "sms.msg_tmpl",
            "Factored: your security code is {code}. Please don't reply.")
        plivo_auth_id = self.req.registry.settings.get("sms.plivo_auth_id", None)
        plivo_auth_token = self.req.registry.settings.get("sms.plivo_auth_token", None)
        plivo_phone_number = self.req.registry.settings.get("sms.plivo_phone_number", None)
        if not plivo_auth_id or not plivo_auth_token or not plivo_phone_number:
            import pdb; pdb.set_trace()
            logger.error("plivo not completely configured, check sms.plivo_* settings")
            raise HTTPInternalServerError()
        if "{code}" not in sms_msg_tmpl:
            logger.error("sms.msg_tmpl is not configured with a {code} variable")
            raise HTTPInternalServerError()

        # generate code for this request
        user.generated_code = make_random_code(6)
        user.generated_code_time_stamp = datetime.utcnow()
        self.db.save(self.req, user)

        # prepare phone numbers
        ## only digits, and there needs to be 11 of them (country, area, phone)
        src_number = ''.join([d for d in plivo_phone_number if d in string.digits])
        dest_number = ''.join([d for d in user.phone if d in string.digits])
        if len(src_number) != 11:
            logger.error("plivo number is incomplete: must include country "
                         "code, area code, and phone number -- 11 digits in total."
                         "This is the number configured: {}".format(src_number))
            raise HTTPInternalServerError()
        if len(dest_number) != 11:
            logger.error("user phone number is incomplete: must include country "
                         "code, area code, and phone number -- 11 digits in total."
                         "This is the number received: {}".format(dest_number))
            raise HTTPInternalServerError()

        # call out to plivo to send message
        plivo_endpoint = "https://api.plivo.com/v1/Account/{}/Message/".format(plivo_auth_id)
        payload = dict(
            src=src_number,
            dst=dest_number,
            text=sms_msg_tmpl.format(code=user.generated_code)
        )
        response = requests.post(plivo_endpoint, json=payload, auth=(plivo_auth_id, plivo_auth_token))

        # handle/log the result
        if response.status_code != 202:
            logger.error("problem sending sms:")
            try:
                logger.error(response.json())
            except:
                logger.error(response.text)
            raise HTTPInternalServerError()
        else:
            rjson = response.json()
            logger.info("User '{}' sent msg to '{}'. plivo msg uuid is '{}'".format(
                user.username,
                dest_number,
                rjson.get("message_uuid", "unknown")))

    def check_code(self, user, code=None):
        try:
            window = int(self.req.registry.settings.get('sms.auth_window', 120))
        except ValueError:
            logger.error("sms.auth_window is not set to a valid integer. "
                         "Setting to default of 120 seconds")
            window = 120
        codetocheck = code if code else self.cform.data['code']
        now = datetime.utcnow()
        expires_at = user.generated_code_time_stamp + timedelta(seconds=window)

        # out of time
        if now >= expires_at:
            return False

        # bad code entered
        if strings_differ(codetocheck, user.generated_code):
            return False

        return True


addFactoredPlugin(SMSAuthPlugin)



from sqlalchemy import engine_from_config
from factored.models import DBSession, User
from sqlalchemy import func
from factored.utils import generate_random_google_code


class DB(object):

    def __init__(self):
        pass

    def get_user(self, req, username):
        db = req.sm[req.registry.settings['db_session_id']]
        return db.query(User).filter(
            func.lower(User.username) == func.lower(username)).first()

    def create_user(self, req, username):
        db = req.sm[req.registry.settings['db_session_id']]
        secret = generate_random_google_code()
        user = User(username=username.lower(), secret=secret)
        db.add(user)
        return user

    def save(self, req, user):
        # not necessary here?
        pass


def factory(settings, app):
    engine = engine_from_config(settings, 'sqlalchemy.')
    configure_db = settings.pop('configure_db', 'true').lower() == 'true'

    if configure_db:
        DBSession.configure(bind=engine)
        db_session_id = 'f'
    else:
        # why would you do this?
        # well... if you have multiple factored wsgi instances running
        # this allows you to change between the connection objects
        db_session_id = settings.pop('db_session_id')
    return db_session_id, DB()



#!/usr/bin/env python
"""A noddy fake smtp server."""

import smtpd
import asyncore
from pprint import pprint


class FakeSMTPServer(smtpd.SMTPServer):
    """A Fake smtp server"""

    def __init__(*args, **kwargs):
        print("Running fake smtp server on port 25")
        smtpd.SMTPServer.__init__(*args, **kwargs)

    def process_message(*args, **kwargs):
        pprint(args)
        pprint(kwargs)

if __name__ == "__main__":
    smtp_server = FakeSMTPServer(('localhost', 25), None)
    try:
        asyncore.loop()
    except KeyboardInterrupt:
        smtp_server.close()



import argparse

from sqlalchemy import engine_from_config

from pyramid.paster import get_appsettings, setup_logging
from factored.models import DBSession, User
from factored.utils import get_barcode_image
from factored.utils import create_user


addparser = argparse.ArgumentParser(description='Add user')
addparser.add_argument('config', help='configuration file')
addparser.add_argument('--username', dest='username', help='username')


def add():
    arguments = addparser.parse_args()
    if not arguments.config or not arguments.username:
        addparser.print_usage()
    else:
        config_uri = arguments.config
        setup_logging(config_uri)
        try:
            settings = get_appsettings(config_uri, 'factored')
        except LookupError:
            settings = get_appsettings(config_uri, 'main')
        engine = engine_from_config(settings, 'sqlalchemy.')
        DBSession.configure(bind=engine)
        session = DBSession()
        username = arguments.username
        user = create_user(username, session)
        print('barcode url:', get_barcode_image(username, user.secret,
                                                settings['appname']))
        print('secret:', user.secret)
        session.commit()
        session.close()

removeparser = argparse.ArgumentParser(description='Remove user')
removeparser.add_argument('config', help='configuration file')
removeparser.add_argument('--username', dest='username', help='username')


def remove():
    arguments = removeparser.parse_args()
    if not arguments.config or not arguments.username:
        removeparser.print_usage()
    else:
        config_uri = arguments.config
        setup_logging(config_uri)
        try:
            settings = get_appsettings(config_uri, 'factored')
        except LookupError:
            settings = get_appsettings(config_uri, 'main')
        engine = engine_from_config(settings, 'sqlalchemy.')
        DBSession.configure(bind=engine)
        session = DBSession()
        user = session.query(User).filter_by(
            username=arguments.username).all()
        if len(user) > 0:
            session.delete(user[0])
        else:
            print('"%s" user not found' % arguments.username)
        session.commit()
        session.close()

listparser = argparse.ArgumentParser(description='Remove user')
listparser.add_argument('config', help='configuration file')


def listusers():
    arguments = removeparser.parse_args()
    if not arguments.config:
        removeparser.print_usage()
    else:
        config_uri = arguments.config
        setup_logging(config_uri)
        try:
            settings = get_appsettings(config_uri, 'factored')
        except LookupError:
            settings = get_appsettings(config_uri, 'main')
        engine = engine_from_config(settings, 'sqlalchemy.')
        DBSession.configure(bind=engine)
        session = DBSession()
        for user in session.query(User).all():
            print(user.username)


listuserparser = argparse.ArgumentParser(description='List user info')
listuserparser.add_argument('config', help='configuration file')
listuserparser.add_argument('--username', dest='username', help='username')


def listuserinfo():
    arguments = listuserparser.parse_args()
    if not arguments.config or not arguments.username:
        listuserparser.print_usage()
    else:
        config_uri = arguments.config
        setup_logging(config_uri)
        try:
            settings = get_appsettings(config_uri, 'factored')
        except LookupError:
            settings = get_appsettings(config_uri, 'main')
        engine = engine_from_config(settings, 'sqlalchemy.')
        DBSession.configure(bind=engine)
        session = DBSession()
        users = session.query(User).filter_by(
            username=arguments.username).all()
        if len(users) > 0:
            user = users[0]
            print('username:%s, secret: %s' % (
                user.username, user.secret))
            print('bar code url:', get_barcode_image(user.username,
                                                     user.secret,
                                                     settings['appname']))
        else:
            print('"%s" user not found' % arguments.username)



import unittest
from pyramid import testing
from factored.models import DBSession
from factored.utils import generate_random_google_code
from datetime import timedelta
from factored.app import Authenticator


class FakeMailer(object):

    def __init__(self):
        self.messages = []

    def send(self, message):
        self.messages.append(message)

    send_immediately = send


def test_application(environ, start_response):
    body = 'foobar'
    headers = [('Content-Type', 'text/html; charset=utf8'),
               ('Content-Length', str(len(body)))]
    start_response('200 Ok', headers)
    return [body]


class FakeApp(object):
    base_auth_url = '/auth'


class BaseTest(unittest.TestCase):

    def tearDown(self):
        self.session.close()
        testing.tearDown()

    def setUp(self):
        self.config = testing.setUp()
        self.settings = {
            'sqlalchemy.url': 'sqlite://',
            'em.subject': 'Authentication Request',
            'em.sender': 'foo@bar.com',
            'em.body': """
    You have requested authentication.
    Your temporary access code is: {code}""",
            'auth_tkt.secret': 'secret',
            'auth_tkt.cookie_name': 'pnutbtr',
            'supported_auth_schemes': [
                'Google Auth',
                'Email'
            ]
        }
        self.app = Authenticator(test_application, {}, **self.settings)
        from webtest import TestApp
        self.testapp = TestApp(self.app)
        self.mailer = FakeMailer()
        self.testapp.app.pyramid.registry['mailer'] = self.mailer
        from sqlalchemy import create_engine
        engine = create_engine('sqlite://')
        from factored.models import Base
        DBSession.configure(bind=engine)
        self.session = DBSession()
        Base.metadata.create_all(engine)


class TestSome(BaseTest):

    def setUp(self):
        super(TestSome, self).setUp()
        from factored.models import User
        self.user = User(
            username='foo@bar.com', secret=generate_random_google_code())
        self.session.add(self.user)
        self.session.commit()

    def test_redirect_to_auth(self):
        self.testapp.get('/', status=302)


class TestGoogleAuth(BaseTest):

    def setUp(self):
        super(TestGoogleAuth, self).setUp()
        from factored.models import User
        self.user = User(username='foo', secret=generate_random_google_code())
        self.session.add(self.user)

    def test_blank_form(self):
        resp = self.testapp.get('/auth/ga', status=200)
        resp.mustcontain('name="username')

    def test_submit_without_username_gives_error(self):
        resp = self.testapp.post('/auth/ga', status=200, params={
            'username': '',
            'submit': 'Next'
        })
        resp.mustcontain('class="error')

    def test_submit_with_wrong_code(self):
        resp = self.testapp.post('/auth/ga', status=200, params={
            'submit': 'Authenticate',
            'username': 'foo',
            'code': '377474'
        })
        resp.mustcontain('class="error')
        resp.mustcontain('Invalid username')

    def test_submit_success_with_code(self):
        from factored.utils import get_google_auth_code
        from factored.models import User
        user = self.session.query(User).filter_by(username='foo').first()
        self.testapp.post('/auth/ga', status=302, params={
            'submit': 'Authenticate',
            'username': 'foo',
            'code': get_google_auth_code(user.secret)
        })
        self.testapp.get('/')

    def test_submit_success_with_code_check_headers(self):
        from factored.utils import get_google_auth_code
        from factored.models import User
        user = self.session.query(User).filter_by(username='foo').first()
        resp = self.testapp.post('/auth/ga', status=302, params={
            'submit': 'Authenticate',
            'username': 'foo',
            'code': get_google_auth_code(user.secret)
        })
        assert 'pnutbtr=' in resp.headers['Set-Cookie']


class TestEmailAuth(BaseTest):

    def setUp(self):
        super(TestEmailAuth, self).setUp()
        from factored.models import User
        self.user = User(
            username='foo@bar.com', secret=generate_random_google_code())
        self.session.add(self.user)
        self.session.commit()

    def test_email_blank_form(self):
        resp = self.testapp.get('/auth/em', status=200)
        resp.mustcontain('name="username')

    def test_not_send_mail_without_username(self):
        resp = self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail'
        })
        resp.mustcontain('class="error')

    def test_not_send_mail_with_incorrect_username_non_email(self):
        resp = self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail',
            'username': 'foobar'
        })
        resp.mustcontain('class="error')
        resp.mustcontain('must contain a single @')

    def test_not_send_mail_with_incorrect_username(self):
        resp = self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail',
            'username': 'blah@foo.com'
        })
        resp.mustcontain('class="error')
        assert 'must contain a single @' not in resp.body
        resp.mustcontain('Should be the same as the username.')

    def test_send_mail_with_correct_username(self):
        from factored.models import User
        resp = self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail',
            'username': 'foo@bar.com'
        })
        user = self.session.query(User).filter_by(
            username='foo@bar.com').first()
        assert 'class="error' not in resp.body
        self.assertTrue(len(self.mailer.messages) == 1)
        self.assertTrue(user.generated_code in self.mailer.messages[0].body)

    def test_send_mail_with_correct_username_case_insensitive(self):
        from factored.models import User
        resp = self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail',
            'username': 'FOO@BAR.com'
        })
        user = self.session.query(User).filter_by(
            username='foo@bar.com').first()
        assert 'class="error' not in resp.body
        self.assertTrue(len(self.mailer.messages) == 1)
        self.assertTrue(user.generated_code in self.mailer.messages[0].body)

    def test_auth_correct(self):
        from factored.models import User
        self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail',
            'username': 'foo@bar.com'
        })
        user = self.session.query(User).filter_by(
            username='foo@bar.com').first()
        self.testapp.post('/auth/em', status=302, params={
            'submit': 'Authenticate',
            'username': 'foo@bar.com',
            'code': user.generated_code
        })
        self.testapp.get('/')

    def test_auth_correct_sets_headers(self):
        from factored.models import User
        self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail',
            'username': 'foo@bar.com'
        })
        user = self.session.query(User).filter_by(
            username='foo@bar.com').first()
        resp = self.testapp.post('/auth/em', status=302, params={
            'submit': 'Authenticate',
            'username': 'foo@bar.com',
            'code': user.generated_code
        })
        assert 'pnutbtr=' in resp.headers['Set-Cookie']

    def test_auth_fails_bad_username(self):
        from factored.models import User
        self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail',
            'username': 'foo@bar.com'
        })
        user = self.session.query(User).filter_by(
            username='foo@bar.com').first()
        resp = self.testapp.post('/auth/em', status=200, params={
            'submit': 'Authenticate',
            'username': 'foo3@bar.com',
            'code': user.generated_code
        })
        resp.mustcontain('class="error')

    def test_auth_fails_missing_code(self):
        self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail',
            'username': 'foo@bar.com'
        })
        resp = self.testapp.post('/auth/em', status=200, params={
            'submit': 'Authenticate',
            'username': 'foo@bar.com'
        })
        resp.mustcontain('class="error')

    def test_auth_fails_bad_code(self):
        self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail',
            'username': 'foo@bar.com'
        })
        resp = self.testapp.post('/auth/em', status=200, params={
            'submit': 'Authenticate',
            'username': 'foo@bar.com',
            'code': 'random'
        })
        resp.mustcontain('class="error')

    def test_auth_fails_time_limit(self):
        from factored.models import User
        self.testapp.post('/auth/em', status=200, params={
            'submit': 'Send mail',
            'username': 'foo@bar.com'
        })
        user = self.session.query(User).filter_by(
            username='foo@bar.com').first()
        user.generated_code_time_stamp = \
            user.generated_code_time_stamp - timedelta(seconds=121)
        self.session.commit()
        resp = self.testapp.post('/auth/em', status=200, params={
            'submit': 'Authenticate',
            'username': 'foo@bar.com',
            'code': user.generated_code
        })
        resp.mustcontain('class="error')



from pyramid.events import subscriber
from pyramid.events import BeforeRender
from factored.templates import TemplateRendererFactory


@subscriber(BeforeRender)
def add_globals(event):
    req = event['request']
    view = event['view']
    context = event.rendering_val
    if context is None:
        return {}
    if 'static_path' not in context:
        context['static_path'] = req.registry['settings']['static_path']
    if 'req' not in context:
        context['req'] = req
    if 'templates' not in context:
        templates = TemplateRendererFactory(req, context)
        context['templates'] = templates

        def render(name):
            if ':' in name:
                return templates.render(name)
            else:
                return templates.render('templates/%s' % name)
        context['render'] = render
    if 'content_renderer' not in context:
        context['content_renderer'] = 'templates/auth-chooser.pt'
    if 'allow_code_reminder' not in context:
        if hasattr(view, 'allow_code_reminder'):
            context['allow_code_reminder'] = view.allow_code_reminder
        elif 'auth_plugin' in context:
            context['allow_code_reminder'] = \
                context['auth_plugin'].allow_code_reminder
    # update app settings as template globals. Only missing ones.
    for key, value in req.registry['settings'].items():
        if key not in context:
            context[key] = value
    return context



from pyramid.request import Request as BaseRequest
from factored.sm import getSessionManager
from pyramid.decorator import reify


class Request(BaseRequest):
    @reify
    def sm(self):
        return getSessionManager(self.environ)



from sqlalchemy import create_engine

import logging
logger = logging.getLogger(__name__)

# enable/disable LDAP user finder plugin
try:
    import ldap
    LDAP_AVAILABLE = True
except ImportError:
    logger.warning("python-ldap module not found, LDAPUserFinderPlugin "
                   "won't be enabled")
    LDAP_AVAILABLE = False




_finder_plugins = []


def addUserFinderPlugin(plugin):
    _finder_plugins.append(plugin)


def getUserFinderPlugins():
    return _finder_plugins


def getUserFinderPlugin(name):
    for plugin in _finder_plugins:
        if plugin.name == name:
            return plugin


class SQLUserFinder(object):
    """
    Auto find related item from a SQL database
    """
    name = "SQL"

    def __init__(self, table_name, email_field, connection_string):
        self.connection_string = connection_string
        self.table_name = table_name
        self.email_field = email_field

    @property
    def engine(self):
        return create_engine(self.connection_string)

    def __call__(self, username):
        select = 'select %s from %s where %s=?' % (
            self.email_field, self.table_name, self.email_field)
        engine = self.engine
        if engine.driver == 'mysqldb':
            select = select.replace('?', '%s')
        res = engine.execute(select, username)
        return len(res.fetchall()) > 0
addUserFinderPlugin(SQLUserFinder)


class EmailDomainFinderPlugin(object):
    name = "Email Domain"

    def __init__(self, valid_domains):
        from factored import app
        self.valid_domains = app._tolist(valid_domains)

    def __call__(self, email):
        email = email.lower()
        for valid in self.valid_domains:
            if email.endswith('@' + valid):
                return True
        return False
addUserFinderPlugin(EmailDomainFinderPlugin)


# Should handle:
#   - LDAPS (or no ldap, can disable certificate checkcs as well)
#   - STARTTLS (or not)
#   - simple bind (bind user DN and password)
#
# All this user finder does is connect, bind with an account that can search
# the specified base_dn, and then search for a _single_ result in which the
# given username matches the value of the username_attr on the object
#
# example values:
#   conn_string: ldaps://example.com
#   starttls: True
#   bind_dn: cn=test,ou=users,dc=example,dc=com
#   bind_pw: abc123
#   base_dn: ou=users,dc=example,dc=com
#   username_attr: mail
#   lookup_timeout: 60
class LDAPUserFinderPlugin(object):
    name = "LDAP Users"

    def __init__(self, conn_string, check_certificate, starttls, bind_dn, bind_pw, base_dn,
                 username_attr, lookup_timeout):
        self.conn_string = conn_string
        self.check_certificate = check_certificate.lower().strip() == "true"
        self.starttls = starttls.lower().strip() == "true"
        self.bind_dn = bind_dn
        self.bind_pw = bind_pw
        self.base_dn = base_dn
        self.username_attr = username_attr
        self.lookup_timeout = lookup_timeout

    @property
    def conn(self):
        if not self.check_certificate:
            ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)

        conn = ldap.initialize(self.conn_string)
        if self.starttls:
            try:
                conn.start_tls_s()
            except ldap.LDAPError as e:
                logger.error("can't STARTTLS")
                if type(e.message) == dict and e.message.has_key("desc"):
                    logger.error(e.message['desc'])
                else:
                    logger.error(e)
                conn.unbind()
                return None
        try:
            conn.simple_bind_s(self.bind_dn, self.bind_pw)
        except ldap.INVALID_CREDENTIALS:
            logger.info("invalid credentials for bind dn: " + self.bind_dn)
            conn.unbind()
            return None
        except ldap.LDAPError as e:
            logger.error("LDAP error when binding")
            if type(e.message) == dict and e.message.has_key('desc'):
                logger.error(e.message['desc'])
            else:
                logger.error(e)
            conn.unbind()
            return None
        return conn

    def __call__(self, username):
        conn = self.conn
        if not conn:
            return False

        found = False
        try:
            userfilter = "({}={})".format(self.username_attr, username)
            results = conn.search_st(
                self.base_dn,
                ldap.SCOPE_SUBTREE,
                userfilter,
                attrlist=[self.username_attr])
            if len(results) > 0:
                if len(results) > 1:
                    logger.error("more than one user found with that username")
                else:
                    found = True
        except ldap.LDAPError as e:
            logger.error("LDAP error when performing user lookup")
            if type(e.message) == dict and e.message.has_key('desc'):
                logger.error(e.message['desc'])
            else:
                logger.error(e)

        # make sure the connection gets unbound
        #conn.unbind()

        return found

# LDAP User Finder is a WIP so disabled for now
if LDAP_AVAILABLE:
    addUserFinderPlugin(LDAPUserFinderPlugin)



from setuptools import setup, find_packages

version = '4.1.0'

requires = [
    'WebOb',
    'pyramid',
    'SQLAlchemy',
    'pyramid_simpleform',
    'pyramid_mailer',
    'argparse',
    'pyramid_tm',
    'pyramid_chameleon'
]

setup(name='factored',
      version=version,
      description="A WSGI app that allows you to add another factor of "
                  "authentication to any application server.",
      long_description="%s\n%s" % (
          open("README.rst").read(),
          open("CHANGES.rst").read()),
      classifiers=[
          'Topic :: Internet :: WWW/HTTP :: WSGI',
          'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
          'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware'],
      keywords='mutli factor authentication 2factor mutl-factor '
               'auth auth_tkt google otp',
      author='Nathan Van Gheem',
      author_email='vangheem@gmail.com',
      url='https://github.com/wildcardcorp/factored',
      license='GPL2',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      tests_require=requires + ['WebTest'],
      extras_require={
          'test': [
              'WebTest',
              'pyramid_debugtoolbar'
          ],
          'proxy': [
              'WSGIProxy'
          ]
      },
      test_suite="factored",
      entry_points={
          'paste.app_factory': [
              'simpleproxy = factored.app:SimpleProxy',
              'main = factored.app:Authenticator'],
          'paste.filter_app_factory': [
              'main = factored.app:Authenticator',
              'sm = factored.sm:make_sm'],
          'console_scripts': [
              'factored_initializedb = factored.scripts.initializedb:main',
              'factored_adduser = factored.scripts.users:add',
              'factored_removeuser = factored.scripts.users:remove',
              'factored_listusers = factored.scripts.users:listusers',
              'factored_listuserinfo = factored.scripts.users:listuserinfo'
          ],
          'factored.db_factory': [
              'sql = factored.sql:factory'
          ]
      })
